/* Egoboo - camera.c
 * Various functions related to how the game camera works.
 */

/*
    This file is part of Egoboo.

    Egoboo is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Egoboo is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Egoboo.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "egoboo.h"
#include "mathstuff.h"
#include <assert.h>

//--------------------------------------------------------------------------------------------
void camera_look_at(float x, float y)
{
  // ZZ> This function makes the camera turn to face the character
  camzgoto = camzadd;
  camturnleftright = (3.0f * PI_OVER_TWO) - atan2(y - camy, x - camx);  // xgg
}

//--------------------------------------------------------------------------------------------
void screen_dump_matrix(GLMATRIX a)
{
  int i, j;
  STRING buffer1 = {0};
  STRING buffer2 = {0};

  reset_messages();
  for (j=0; j<4; j++)
  {
    snprintf(buffer1, sizeof(buffer1), "  ");
    for (i=0; i<4; i++)
    {
      snprintf(buffer2, sizeof(buffer2), "%2.4f ", (a)_CNV(i, j));
      strncat(buffer1, buffer2, sizeof(buffer1));
    };
    debug_message(1, buffer1);
    buffer1[0] = 0x00;
  }
}

//--------------------------------------------------------------------------------------------
void stdout_dump_matrix(GLMATRIX a)
{
  int i, j;

  for (j=0; j<4; j++)
  {
    printf("  ");
    for (i=0; i<4; i++)
    {
      printf("%2.4f ", (a)_CNV(i, j));
    };
    printf("\n");
  }
  printf("\n");
}

//--------------------------------------------------------------------------------------------
void project_view()
{
  // ZZ> This function figures out where the corners of the view area
  //     go when projected onto the plane of the mesh.  Used later for
  //     determining which mesh fans need to be rendered

  int cnt, tnc, extra[3];
  float ztemp;
  float numstep;
  float zproject;
  float xfin, yfin, zfin;
  GLMATRIX mTemp;

  // Range
  ztemp = (camz);

  // Topleft
  //printf("DIAG: In project_view\n");
  //printf("DIAG: dumping mView\n"); dump_matrix(mView);
  //printf("cam xyz,zoom = %f %f %f %f\n",camx,camy,camz,camzoom);

  mTemp = MatrixMult(RotateY(-rotmeshtopside * 0.5f * DEG_TO_RAD), mView);
  mTemp = MatrixMult(RotateX(rotmeshup * 0.5f * DEG_TO_RAD), mTemp);
  zproject = (mTemp)_CNV(2, 2);        //2,2
  // Camera must look down
  if (zproject < 0)
  {
    numstep = -ztemp / zproject;
    xfin = camx + (numstep * (mTemp)_CNV(0, 2));  // xgg   //0,2
    yfin = camy + (numstep * (mTemp)_CNV(1, 2));     //1,2
    zfin = 0;
    cornerx[0] = xfin;
    cornery[0] = yfin;
    //printf("Camera TL: %f %f\n",xfin,yfin);
    //dump_matrix(mTemp);
  }

  // Topright
  mTemp = MatrixMult(RotateY(rotmeshtopside * 0.5f * DEG_TO_RAD), mView);
  mTemp = MatrixMult(RotateX(rotmeshup * 0.5f * DEG_TO_RAD), mTemp);
  zproject = (mTemp)_CNV(2, 2);        //2,2
  // Camera must look down
  if (zproject < 0)
  {
    numstep = -ztemp / zproject;
    xfin = camx + (numstep * (mTemp)_CNV(0, 2));  // xgg   //0,2
    yfin = camy + (numstep * (mTemp)_CNV(1, 2));     //1,2
    zfin = 0;
    cornerx[1] = xfin;
    cornery[1] = yfin;
    //printf("Camera TR: %f %f\n",xfin,yfin);
    //dump_matrix(mTemp);
  }

  // Bottomright
  mTemp = MatrixMult(RotateY(rotmeshbottomside * 0.5f * DEG_TO_RAD), mView);
  mTemp = MatrixMult(RotateX(-rotmeshdown * 0.5f * DEG_TO_RAD), mTemp);
  zproject = (mTemp)_CNV(2, 2);        //2,2
  // Camera must look down
  if (zproject < 0)
  {
    numstep = -ztemp / zproject;
    xfin = camx + (numstep * (mTemp)_CNV(0, 2));  // xgg   //0,2
    yfin = camy + (numstep * (mTemp)_CNV(1, 2));     //1,2
    zfin = 0;
    cornerx[2] = xfin;
    cornery[2] = yfin;
    //printf("Camera BR: %f %f\n",xfin,yfin);
    //dump_matrix(mTemp);
  }

  // Bottomleft
  mTemp = MatrixMult(RotateY(-rotmeshbottomside * 0.5f * DEG_TO_RAD), mView);
  mTemp = MatrixMult(RotateX(-rotmeshdown * 0.5f * DEG_TO_RAD), mTemp);
  zproject = (mTemp)_CNV(2, 2);        //2,2
  // Camera must look down
  if (zproject < 0)
  {
    numstep = -ztemp / zproject;
    xfin = camx + (numstep * (mTemp)_CNV(0, 2));  // xgg   //0,2
    yfin = camy + (numstep * (mTemp)_CNV(1, 2));     //1,2
    zfin = 0;
    cornerx[3] = xfin;
    cornery[3] = yfin;
    //printf("Camera BL: %f %f\n",xfin,yfin);
    //dump_matrix(mTemp);
  }

  // Get the extreme values
  cornerlowx = cornerx[0];
  cornerlowy = cornery[0];
  cornerhighx = cornerx[0];
  cornerhighy = cornery[0];
  cornerlistlowtohighy[0] = 0;
  cornerlistlowtohighy[3] = 0;

  for (cnt = 0; cnt < 4; cnt++)
  {
    if (cornerx[cnt] < cornerlowx)
      cornerlowx = cornerx[cnt];
    if (cornery[cnt] < cornerlowy)
    {
      cornerlowy = cornery[cnt];
      cornerlistlowtohighy[0] = cnt;
    }
    if (cornerx[cnt] > cornerhighx)
      cornerhighx = cornerx[cnt];
    if (cornery[cnt] > cornerhighy)
    {
      cornerhighy = cornery[cnt];
      cornerlistlowtohighy[3] = cnt;
    }
  }

  // Figure out the order of points
  tnc = 0;
  for (cnt = 0; cnt < 4; cnt++)
  {
    if (cnt != cornerlistlowtohighy[0] && cnt != cornerlistlowtohighy[3])
    {
      extra[tnc] = cnt;
      tnc++;
    }
  }
  cornerlistlowtohighy[1] = extra[1];
  cornerlistlowtohighy[2] = extra[0];
  if (cornery[extra[0]] < cornery[extra[1]])
  {
    cornerlistlowtohighy[1] = extra[0];
    cornerlistlowtohighy[2] = extra[1];
  }

  // BAD: exit here
  //printf("Corners:\n");
  //printf("x: %d %d\n",cornerlowx,cornerhighx);
  //printf("y: %d %d\n",cornerlowy,cornerhighy);
  /*printf("Exiting, camera code is broken\n");
  exit(0);*/
}

//--------------------------------------------------------------------------------------------
void make_camera_matrix()
{
  // ZZ> This function sets mView to the camera's location and rotation
  mView = mViewSave;

  mView = MatrixMult(Translate(camx, -camy, camz), mView);  // xgg
  mView = MatrixMult(RotateZ(camturnleftright), mView);
  mView = MatrixMult(RotateX(camturnupdown),    mView);

  if (camswingamp > 0)
  {
    camroll = turntosin[camswing & TRIGTABLE_MASK] * camswingamp * 5;
    mView = MatrixMult(RotateZ(camroll), mView);
  }

}

//--------------------------------------------------------------------------------------------
void bound_camera()
{
  // ZZ> This function stops the camera from moving off the mesh
  if (camx < EDGE)  camx = EDGE;
  if (camx > meshedgex - EDGE)  camx = meshedgex - EDGE;
  if (camy < EDGE)  camy = EDGE;
  if (camy > meshedgey - EDGE)  camy = meshedgey - EDGE;
}

//--------------------------------------------------------------------------------------------
void bound_camtrack()
{
  // ZZ> This function stops the camera target from moving off the mesh
  if (usefaredge)
  {
    if (camtrackx < FARTRACK)  camtrackx = FARTRACK;
    if (camtrackx > meshedgex - FARTRACK)  camtrackx = meshedgex - FARTRACK;
    if (camtracky < FARTRACK)  camtracky = FARTRACK;
    if (camtracky > meshedgey - FARTRACK)  camtracky = meshedgey - FARTRACK;
  }
  else
  {
    if (camtrackx < EDGETRACK)  camtrackx = EDGETRACK;
    if (camtrackx > meshedgex - EDGETRACK)  camtrackx = meshedgex - EDGETRACK;
    if (camtracky < EDGETRACK)  camtracky = EDGETRACK;
    if (camtracky > meshedgey - EDGETRACK)  camtracky = meshedgey - EDGETRACK;
  }
}

//--------------------------------------------------------------------------------------------
void adjust_camera_angle(int height)
{
  // ZZ> This function makes the camera look downwards as it is raised up
  float percentmin, percentmax;


  if (height < MINZADD)  height = MINZADD;
  percentmax = (height - MINZADD) / (float)(MAXZADD - MINZADD);
  percentmin = 1.0 - percentmax;

  camturnupdown = ((MINUPDOWN * percentmin) + (MAXUPDOWN * percentmax));
  camzoom = (MINZOOM * percentmin) + (MAXZOOM * percentmax);
}

//--------------------------------------------------------------------------------------------
void move_camera(float dUpdate)
{
  // ZZ> This function moves the camera
  int cnt, locoalive, movex, movey;  //Used in rts remove? -> int band,
  float x, y, z;
  float vx,vy,vz;
  float level;
  Uint16 character, turnsin, turncos;

  float fkeep, fnew;

  fkeep = pow(camsustain, dUpdate);
  fnew  = 1.0-fkeep;

  if (CData.autoturncamera)
    doturntime = 255;
  else if (doturntime > dUpdate)
  {
    doturntime -= dUpdate;
  }
  else
  {
    doturntime = 0;
  }

  x = y = z = 0;
  vx =vy = vz = 0;
  level = 0;
  locoalive = 0;
  for (cnt = 0; cnt < MAXPLAYER; cnt++)
  {
    if (plavalid[cnt] && pladevice[cnt] != INPUTNONE)
    {
      character = plaindex[cnt];
      if (chralive[character])
      {
        if (chrattachedto[character] == MAXCHR)
        {
          // The character is on foot
          x += chrxpos[character];
          y += chrypos[character];
          z += chrzpos[character];
          level += chrlevel[character];
          vx += chrxvel[character];
          vy += chryvel[character];
          vz += chrzvel[character];
        }
        else
        {
          // The character is mounted
          x += chrxpos[chrattachedto[character]];
          y += chrypos[chrattachedto[character]];
          z += chrzpos[chrattachedto[character]];
          level += chrlevel[chrattachedto[character]];

          vx += chrxvel[chrattachedto[character]];
          vy += chryvel[chrattachedto[character]];
          vz += chrzvel[chrattachedto[character]];
        }
        locoalive++;
      }
    }
  }

  if (locoalive == 0)
  {
    x = camtrackx;
    y = camtracky;
    z = camtrackz;
    vx = camtrackxvel;
    vy = camtrackyvel;
    vz = camtrackzvel;
  }
  else if (locoalive > 1)
  {
    x /= locoalive;
    y /= locoalive;
    z /= locoalive;
    level /= locoalive;
    vx /= locoalive;
    vy /= locoalive;
    vz /= locoalive;
  }

  camtrackxvel = camtrackx * fkeep + vx * fnew;
  camtrackyvel = camtracky * fkeep + vy * fnew;
  camtrackzvel = camtrackz * fkeep + vz * fnew;

  camtrackx = camtrackx * fkeep + x * fnew;
  camtracky = camtracky * fkeep + y * fnew;
  camtrackz = camtrackz * fkeep + z * fnew;

  camtracklevel = camtracklevel * fkeep + level * fnew;
  camzadd       = camzadd  * fkeep + camzaddgoto * fnew;
  camz          = camz * fkeep + camzgoto * fnew;
  camturnadd   *= fkeep;

  // Camera controls
  if (CData.autoturncamera == 255 && numlocalpla >= 1)
  {
    if (mouseon && !control_mouse_is_pressed(MOS_CAMERA))
        camturnadd -= (mousedx * .5) * dUpdate / numlocalpla;

    if (keyon)
    {
      if(control_key_is_pressed(KEY_CAMERA_LEFT))
        camturnadd += CAMKEYTURN * dUpdate / numlocalpla;

      if (control_key_is_pressed(KEY_CAMERA_RIGHT))
        camturnadd -= CAMKEYTURN * dUpdate / numlocalpla;
    };


    if (joyaon && !control_joya_is_pressed(JOA_CAMERA))
        camturnadd -= joyax * CAMJOYTURN * dUpdate / numlocalpla;

    if (joybon && !control_joyb_is_pressed(JOB_CAMERA))
        camturnadd -= joybx * CAMJOYTURN * dUpdate / numlocalpla;
  }
  
  if(numlocalpla >= 1)
  {
    if (mouseon)
    {
      if (control_mouse_is_pressed(MOS_CAMERA))
      {
        camturnadd += (mousedx / 3.0) * dUpdate / numlocalpla;
        camzaddgoto += (float) mousedy / 3.0 * dUpdate / numlocalpla;
        if (camzaddgoto < MINZADD)  camzaddgoto = MINZADD;
        if (camzaddgoto > MAXZADD)  camzaddgoto = MAXZADD;
        doturntime = TURNTIME;  // Sticky turn...
      }
    }

    // JoyA camera controls
    if (joyaon)
    {
      if (control_joya_is_pressed(JOA_CAMERA))
      {
        camturnadd  += joyax * CAMJOYTURN * dUpdate / numlocalpla;
        camzaddgoto += joyay * CAMJOYTURN * dUpdate / numlocalpla;
        if (camzaddgoto < MINZADD)  camzaddgoto = MINZADD;
        if (camzaddgoto > MAXZADD)  camzaddgoto = MAXZADD;
        doturntime = TURNTIME;  // Sticky turn...
      }
    }

    // JoyB camera controls
    if (joybon)
    {
      if (control_joyb_is_pressed(JOB_CAMERA))
      {
        camturnadd  += joybx * CAMJOYTURN * dUpdate / numlocalpla;
        camzaddgoto += joyby * CAMJOYTURN * dUpdate / numlocalpla;
        if (camzaddgoto < MINZADD)  camzaddgoto = MINZADD;
        if (camzaddgoto > MAXZADD)  camzaddgoto = MAXZADD;
        doturntime = TURNTIME;  // Sticky turn...
      }
    }

    // Keyboard camera controls
    if (keyon)
    {
      if (control_key_is_pressed(KEY_CAMERA_LEFT))
      {
        camturnadd += CAMKEYTURN * dUpdate / numlocalpla;
        doturntime = TURNTIME;  // Sticky turn...
      }

      if (control_key_is_pressed(KEY_CAMERA_RIGHT))
      {
        camturnadd -= CAMKEYTURN * dUpdate / numlocalpla;
        doturntime = TURNTIME;  // Sticky turn...
      }

      if (control_key_is_pressed(KEY_CAMERA_IN))
      {
        camzaddgoto -= CAMKEYTURN * dUpdate / numlocalpla;
      }

      if (control_key_is_pressed(KEY_CAMERA_OUT))
      {
        camzaddgoto += CAMKEYTURN * dUpdate / numlocalpla;
      }

      if (camzaddgoto < MINZADD)  camzaddgoto = MINZADD;
      if (camzaddgoto > MAXZADD)  camzaddgoto = MAXZADD;
    }
  }

  //camx -= (float)((mView)_CNV(0, 0)) * camturnadd;  // xgg
  //camy -= (float)((mView)_CNV(1, 0)) * camturnadd;

  // Make it not break...
  //bound_camtrack();
  //bound_camera();

  // Do distance effects for overlay and background
  if (CData.overlayon)
  {
    // Do fg distance effect
    waterlayeru[0] += vx * waterlayerdistx[0] * dUpdate;
    waterlayerv[0] += vy * waterlayerdisty[0] * dUpdate;
  }

  if (!clearson)
  {
    // Do bg distance effect
    waterlayeru[1] += vx * waterlayerdistx[1] * dUpdate;
    waterlayerv[1] += vy * waterlayerdisty[1] * dUpdate;
  }

  // Get ready to scroll...
  movex = camx - camtrackx;
  movey = camy - camtracky;
  camx = sqrt(movex*movex + movey*movey);
  if(camx > 0)
  {
    movex *= camzoom/camx;
    movey *= camzoom/camx;
  }
  else
  {
    movex = camzoom;
    movey = 0;
  }
  turnsin = ((Uint16)camturnadd*10) & TRIGTABLE_MASK;
  turncos = (turnsin + TRIGTABLE_SHIFT) & TRIGTABLE_MASK;
  camx = (movex * turntosin[turncos]  + movey * turntosin[turnsin]) + camtrackx;
  camy = (-movex * turntosin[turnsin] + movey * turntosin[turncos]) + camtracky;

  // Finish up the camera
  camera_look_at(camtrackx, camtracky);

  //bound_camera();
//        adjust_camera_angle(camz-camtracklevel);
  adjust_camera_angle(camz);

  make_camera_matrix();
}

//--------------------------------------------------------------------------------------------
void reset_camera()
{
  // ZZ> This function makes sure the camera starts in a suitable position
  GLVECTOR t1 = {0, 0, 0};
  GLVECTOR t2 = {0, 0, -1};
  GLVECTOR t3 = {0, 1, 0};

  int cnt, save;
//    int mi;

  camswing = 0;
  camx = meshedgex / 2;
  camy = meshedgey / 2;
  camz = 800;
  camzoom = 1000;
//    rtsx = 0;
//    rtsy = 0;
  camtrackxvel = 0;
  camtrackyvel = 0;
  camtrackzvel = 0;
  camcenterx = camx;
  camcentery = camy;
  camtrackx = camx;
  camtracky = camy;
  camtrackz = 0;
  camturnadd = 0;
  camtracklevel = 0;
  camzadd = 800;
  camzaddgoto = 800;
  camzgoto = 800;
  camturnleftright = (float)(-PI_OVER_FOUR);
  camturnleftrightone = (float)(-PI_OVER_FOUR) / (TWO_PI);
  camturnleftrightshort = 0;
  camturnupdown = (float)(PI_OVER_FOUR);
  camroll = 0;

  // Matrix init stuff (from remove.c)
  rotmeshtopside = ((float)CData.scrx / CData.scry) * ROTMESHTOPSIDE / (1.33333);
  rotmeshbottomside = ((float)CData.scrx / CData.scry) * ROTMESHBOTTOMSIDE / (1.33333);
  rotmeshup = ((float)CData.scrx / CData.scry) * ROTMESHUP / (1.33333);
  rotmeshdown = ((float)CData.scrx / CData.scry) * ROTMESHDOWN / (1.33333);
  mWorld = IdentityMatrix();
  mViewSave = mView = ViewMatrix(t1, t2, t3, 0);
  mProjection = ProjectionMatrix(.001f, 2000.0f, (float)(FOV * DEG_TO_RAD)); // 60 degree FOV
  mProjection = MatrixMult(Translate(0, 0, -.999996), mProjection); // Fix Z value...
  mProjection = MatrixMult(ScaleXYZ(-1, -1, 100000), mProjection);  // HUK // ...'cause it needs it

  //[claforte] Fudge the values.
  mProjection.v[10] /= 2.0;
  mProjection.v[11] /= 2.0;

  save = CData.autoturncamera;
  CData.autoturncamera = btrue;

  for (cnt = 0; cnt < 32; cnt++)
  {
    move_camera(1.0);
    camcenterx = camtrackx;
    camcentery = camtracky;
  }

  CData.autoturncamera = save;
  doturntime = 0;
}

