/* Egoboo - enchant.c
*/

/*
This file is part of Egoboo.

Egoboo is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Egoboo is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with Egoboo.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "egoboo.h"

//--------------------------------------------------------------------------------------------
void do_enchant_spawn(float dUpdate)
{
  // ZZ> This function lets enchantments spawn particles
  int cnt, tnc;
  Uint16 facing, particle, eve, character;


  cnt = 0;
  while (cnt < MAXENCHANT)
  {
    if (encon[cnt])
    {
      eve = enceve[cnt];
      if (evecontspawnamount[eve] > 0)
      {
        encspawntime[cnt] -= dUpdate;
        if (encspawntime[cnt] <= 0) encspawntime[cnt] = 0;

        if (encspawntime[cnt] == 0)
        {
          character = enctarget[cnt];
          encspawntime[cnt] = evecontspawntime[eve];
          facing = chrturnleftright[character];
          tnc = 0;
          while (tnc < evecontspawnamount[eve])
          {
            particle = spawn_one_particle(chrxpos[character], chrypos[character], chrzpos[character],
                                          facing, eve, evecontspawnpip[eve],
                                          MAXCHR, SPAWNLAST, chrteam[encowner[cnt]], encowner[cnt], tnc, MAXCHR);
            facing += evecontspawnfacingadd[eve];
            tnc++;
          }
        }
      }
    }
    cnt++;
  }
}


//--------------------------------------------------------------------------------------------
void disenchant_character(Uint16 cnt)
{
  // ZZ> This function removes all enchantments from a character
  while (chrfirstenchant[cnt] != MAXENCHANT)
  {
    remove_enchant(chrfirstenchant[cnt]);
  }
}

//char.c
//--------------------------------------------------------------------------------------------
void damage_character(Uint16 character, Uint16 direction,
                      int damagebase, int damagerand, Uint8 damagetype, Uint8 team,
                      Uint16 attacker, Uint16 effects)
{
  // ZZ> This function calculates and applies damage to a character.  It also
  //     sets alerts and begins actions.  Blocking and frame invincibility
  //     are done here too.  Direction is 0 if the attack is coming head on,
  //     16384 if from the right, 32768 if from the back, 49152 if from the
  //     left.
  int tnc;
  Uint16 action;
  int damage, basedamage;
  Uint16 experience, model, left, right;


  if (chralive[character] && damagebase >= 0 && damagerand >= 1)
  {
    // Lessen damage for resistance, 0 = Weakness, 1 = Normal, 2 = Resist, 3 = Big Resist
    // This can also be used to lessen effectiveness of healing
    damage = damagebase + (rand() % damagerand);
    basedamage = damage;
    damage >>= (chrdamagemodifier[character][damagetype] & DAMAGESHIFT);


    // Allow charging (Invert damage to mana)
    if (chrdamagemodifier[character][damagetype]&DAMAGECHARGE)
    {
      chrmana[character] += damage;
      if (chrmana[character] > chrmanamax[character])
      {
        chrmana[character] = chrmanamax[character];
      }
      return;
    }


    // Invert damage to heal
    if (chrdamagemodifier[character][damagetype]&DAMAGEINVERT)
      damage = -damage;


    // Remember the damage type
    chrdamagetypelast[character] = damagetype;
    chrdirectionlast[character] = direction;


    // Do it already
    if (damage > 0)
    {
      // Only damage if not invincible
      if (chrdamagetime[character] == 0 && !chrinvictus[character])
      {
        model = chrmodel[character];
        if (0 == (effects&DAMFXBLOC))
        {
          // Only damage if hitting from proper direction
          if (madframefx[chrframe[character]]&MADFXINVICTUS)
          {
            // I Frame...
            direction -= capiframefacing[model];
            left = (~capiframeangle[model]);
            right = capiframeangle[model];
            // Check for shield
            if (chraction[character] >= ACTIONPA && chraction[character] <= ACTIONPD)
            {
              // Using a shield?
              if (chraction[character] < ACTIONPC)
              {
                // Check left hand
                if (chrholdingwhich[character][0] != MAXCHR)
                {
                  left = (~capiframeangle[chrmodel[chrholdingwhich[character][0]]]);
                  right = capiframeangle[chrmodel[chrholdingwhich[character][0]]];
                }
              }
              else
              {
                // Check right hand
                if (chrholdingwhich[character][1] != MAXCHR)
                {
                  left = (~capiframeangle[chrmodel[chrholdingwhich[character][1]]]);
                  right = capiframeangle[chrmodel[chrholdingwhich[character][1]]];
                }
              }
            }
          }
          else
          {
            // N Frame
            direction -= capnframefacing[model];
            left = (~capnframeangle[model]);
            right = capnframeangle[model];
          }
          // Check that direction
          if (direction > left || direction < right)
          {
            damage = 0;
          }
        }



        if (damage != 0)
        {
          if (effects&DAMFXARMO)
          {
            chrlife[character] -= damage;
          }
          else
          {
            chrlife[character] -= ((damage * chrdefense[character]) >> 8);
          }


          if (basedamage > MINDAMAGE)
          {
            // Call for help if below 1/2 life
            //if (chrlife[character] < (chrlifemax[character] >> 1))	//Zefz: Removed, because it caused guards to attack
            //  call_for_help(character);								//when dispelling overlay spells (Faerie Light)

            // Spawn blud particles
            if (capbludvalid[model] && (damagetype < DAMAGEHOLY || capbludvalid[model] == ULTRABLOODY))
            {
              spawn_one_particle(chrxpos[character], chrypos[character], chrzpos[character],
                                 chrturnleftright[character] + direction, chrmodel[character], capbludprttype[model],
                                 MAXCHR, SPAWNLAST, chrteam[character], character, 0, MAXCHR);
            }
            // Set attack alert if it wasn't an accident
            if (team == DAMAGETEAM)
            {
              chrattacklast[character] = MAXCHR;
            }
            else
            {
              // Don't alert the character too much if under constant fire
              if (chrcarefultime[character] == 0)
              {
                // Don't let characters chase themselves...  That would be silly
                if (attacker != character)
                {
                  chralert[character] |= ALERTIFATTACKED;
                  chrattacklast[character] = attacker;
                  chrcarefultime[character] = CAREFULTIME;
                }
              }
            }
          }


          // Taking damage action
          action = ACTIONHA;
          if (chrlife[character] < 0)
          {
            // Character has died
            chralive[character] = bfalse;
            disenchant_character(character);
            //chrwaskilled[character] = btrue;
            chrkeepaction[character] = btrue;
            chrlife[character] = -1;
            chrplatform[character] = btrue;
            chrbumpdampen[character] /= 2.0;
            action = ACTIONKA;
            // Give kill experience
            experience = capexperienceworth[model] + (chrexperience[character] * capexperienceexchange[model]);
            if (attacker < MAXCHR)
            {
              // Set target
              chraitarget[character] = attacker;
              if (team == DAMAGETEAM)  chraitarget[character] = character;
              if (team == NULLTEAM)  chraitarget[character] = character;
              // Award direct kill experience
              if (teamhatesteam[chrteam[attacker]][chrteam[character]])
              {
                give_experience(attacker, experience, XPKILLENEMY);
              }
              // Check for hated
              if (capidsz[chrmodel[attacker]][IDSZHATE] == capidsz[model][IDSZPARENT] ||
                  capidsz[chrmodel[attacker]][IDSZHATE] == capidsz[model][IDSZTYPE])
              {
                give_experience(attacker, experience, XPKILLHATED);
              }
            }
            // Clear all shop passages that it owned...
            tnc = 0;
            while (tnc < numshoppassage)
            {
              if (shopowner[tnc] == character)
              {
                shopowner[tnc] = NOOWNER;
              }
              tnc++;
            }
            // Let the other characters know it died
            tnc = 0;
            while (tnc < MAXCHR)
            {
              if (chron[tnc] && chralive[tnc])
              {
                if (chraitarget[tnc] == character)
                {
                  chralert[tnc] |= ALERTIFTARGETKILLED;
                }
                if (!teamhatesteam[chrteam[tnc]][team] && teamhatesteam[chrteam[tnc]][chrteam[character]])
                {
                  // All allies get team experience, but only if they also hate the dead guy's team
                  give_experience(tnc, experience, XPTEAMKILL);
                }
              }
              tnc++;
            }
            // Check if it was a leader
            if (teamleader[chrteam[character]] == character)
            {
              // It was a leader, so set more alerts
              tnc = 0;
              while (tnc < MAXCHR)
              {
                if (chron[tnc] && chrteam[tnc] == chrteam[character])
                {
                  // All folks on the leaders team get the alert
                  chralert[tnc] |= ALERTIFLEADERKILLED;
                }
                tnc++;
              }
              // The team now has no leader
              teamleader[chrteam[character]] = NOLEADER;
            }
            detach_character_from_mount(character, btrue, bfalse);
            action += (rand() & 3);
            play_action(character, action, bfalse);
            // Turn off all sounds if it's a player
            if (chrisplayer[character])
            {
              tnc = 0;
              while (tnc < MAXWAVE)
              {
                //stop_sound(capwaveindex[chrmodel[character]][tnc]);
                //TODO Zefz: Do we need this? This makes all sounds a character makes stop when it dies...
                tnc++;
              }
            }
            // Afford it one last thought if it's an AI
            teammorale[chrbaseteam[character]]--;
            chrteam[character] = chrbaseteam[character];
            chralert[character] = ALERTIFKILLED;
            chrsparkle[character] = NOSPARKLE;
            chraitime[character] = 1;  // No timeout...
            let_character_think(character);
          }
          else
          {
            if (basedamage > MINDAMAGE)
            {
              action += (rand() & 3);
              play_action(character, action, bfalse);
              // Make the character invincible for a limited time only
              if (!(effects & DAMFXTIME))
                chrdamagetime[character] = DAMAGETIME;
            }
          }
        }
        else
        {
          // Spawn a defend particle
          spawn_one_particle(chrxpos[character], chrypos[character], chrzpos[character], chrturnleftright[character], MAXMODEL, PRTPIP_DEFEND, MAXCHR, SPAWNLAST, NULLTEAM, MAXCHR, 0, MAXCHR);
          chrdamagetime[character] = DEFENDTIME;
          chralert[character] |= ALERTIFBLOCKED;
        }
      }
    }
    else if (damage < 0)
    {
      chrlife[character] -= damage;
      if (chrlife[character] > chrlifemax[character])  chrlife[character] = chrlifemax[character];

      // Isssue an alert
      chralert[character] |= ALERTIFHEALED;
      chrattacklast[character] = attacker;
      if (team != DAMAGETEAM)
      {
        chrattacklast[character] = MAXCHR;
      }
    }
  }
}

//--------------------------------------------------------------------------------------------
void kill_character(Uint16 character, Uint16 killer)
{
  // ZZ> This function kills a character...  MAXCHR killer for accidental death
  Uint8 modifier;

  if (chralive[character])
  {
    chrdamagetime[character] = 0;
    chrlife[character] = 1;
    modifier = chrdamagemodifier[character][DAMAGECRUSH];
    chrdamagemodifier[character][DAMAGECRUSH] = 1;
    if (killer != MAXCHR)
    {
      damage_character(character, 0, 512, 1, DAMAGECRUSH, chrteam[killer], killer, DAMFXARMO | DAMFXBLOC);
    }
    else
    {
      damage_character(character, 0, 512, 1, DAMAGECRUSH, DAMAGETEAM, chrbumplast[character], DAMFXARMO | DAMFXBLOC);
    }
    chrdamagemodifier[character][DAMAGECRUSH] = modifier;
  }
}

//--------------------------------------------------------------------------------------------
void spawn_poof(Uint16 character, Uint16 profile)
{
  // ZZ> This function spawns a character poof
  Uint16 sTmp;
  Uint16 origin;
  int iTmp;


  sTmp = chrturnleftright[character];
  iTmp = 0;
  origin = chraiowner[character];
  while (iTmp < capgopoofprtamount[profile])
  {
    spawn_one_particle(chroldx[character], chroldy[character], chroldz[character],
                       sTmp, profile, capgopoofprttype[profile],
                       MAXCHR, SPAWNLAST, chrteam[character], origin, iTmp, MAXCHR);
    sTmp += capgopoofprtfacingadd[profile];
    iTmp++;
  }
}

//--------------------------------------------------------------------------------------------
void naming_names(int profile)
{
  // ZZ> This function generates a random name
  int read, write, section, mychop;
  char cTmp;

  if (capsectionsize[profile][0] == 0)
  {
    namingnames[0] = 'B';
    namingnames[1] = 'l';
    namingnames[2] = 'a';
    namingnames[3] = 'h';
    namingnames[4] = 0;
  }
  else
  {
    write = 0;
    section = 0;
    while (section < MAXSECTION)
    {
      if (capsectionsize[profile][section] != 0)
      {
        mychop = capsectionstart[profile][section] + (rand() % capsectionsize[profile][section]);
        read = chopstart[mychop];
        cTmp = chopdata[read];
        while (cTmp != 0 && write < MAXCAPNAMESIZE - 1)
        {
          namingnames[write] = cTmp;
          write++;
          read++;
          cTmp = chopdata[read];
        }
      }
      section++;
    }
    if (write >= MAXCAPNAMESIZE) write = MAXCAPNAMESIZE - 1;
    namingnames[write] = 0;
  }
}

//--------------------------------------------------------------------------------------------
void read_naming(int profile, char *szLoadname)
{
  // ZZ> This function reads a naming file
  FILE *fileread;
  int section, chopinsection, cnt;
  char mychop[32], cTmp;

  fileread = fs_fileOpen(PRI_NONE, NULL, szLoadname, "r");
  if (NULL != fileread)
  {
    section = 0;
    chopinsection = 0;
    while (goto_colon_yesno(fileread) && section < MAXSECTION)
    {
      fscanf(fileread, "%s", mychop);
      if (mychop[0] != 'S' || mychop[1] != 'T' || mychop[2] != 'O' || mychop[3] != 'P')
      {
        if (chopwrite >= CHOPDATACHUNK)  chopwrite = CHOPDATACHUNK - 1;
        chopstart[numchop] = chopwrite;
        cnt = 0;
        cTmp = mychop[0];
        while (cTmp != 0 && cnt < 31 && chopwrite < CHOPDATACHUNK)
        {
          if (cTmp == '_') cTmp = ' ';
          chopdata[chopwrite] = cTmp;
          cnt++;
          chopwrite++;
          cTmp = mychop[cnt];
        }
        if (chopwrite >= CHOPDATACHUNK)  chopwrite = CHOPDATACHUNK - 1;
        chopdata[chopwrite] = 0;  chopwrite++;
        chopinsection++;
        numchop++;
      }
      else
      {
        capsectionsize[profile][section] = chopinsection;
        capsectionstart[profile][section] = numchop - chopinsection;
        section++;
        chopinsection = 0;
      }
    }
    fs_fileClose(fileread);
  }
}

//--------------------------------------------------------------------------------------------
void prime_names(void)
{
  // ZZ> This function prepares the name chopper for use
  int cnt, tnc;

  numchop = 0;
  chopwrite = 0;
  cnt = 0;
  while (cnt < MAXMODEL)
  {
    tnc = 0;
    while (tnc < MAXSECTION)
    {
      capsectionstart[cnt][tnc] = MAXCHOP;
      capsectionsize[cnt][tnc] = 0;
      tnc++;
    }
    cnt++;
  }
}

//--------------------------------------------------------------------------------------------
void tilt_characters_to_terrain()
{
  // ZZ> This function sets all of the character's starting tilt values
  int cnt;
  Uint8 twist;

  cnt = 0;
  while (cnt < MAXCHR)
  {
    if (chrstickybutt[cnt] && chron[cnt] && chronwhichfan[cnt] != OFFEDGE)
    {
      twist = meshtwist[chronwhichfan[cnt]];
      chrturnmaplr[cnt] = maplrtwist[twist];
      chrturnmapud[cnt] = mapudtwist[twist];
    }
    cnt++;
  }
}

//--------------------------------------------------------------------------------------------
int spawn_one_character(float x, float y, float z, int profile, Uint8 team,
                        Uint8 skin, Uint16 facing, char *name, int override)
{
  // ZZ> This function spawns a character and returns the character's index number
  //     if it worked, MAXCHR otherwise
  int cnt, tnc;


  // Make sure the team is valid
  if (team > MAXTEAM - 1)
    team = MAXTEAM - 1;


  // Get a new character
  cnt = MAXCHR;
  if (madused[profile])
  {
    if (override < MAXCHR)
    {
      cnt = get_free_character();
      if (cnt != override)
      {
        // Picked the wrong one, so put this one back and find the right one
        tnc = 0;
        while (tnc < MAXCHR)
        {
          if (freechrlist[tnc] == override)
          {
            freechrlist[tnc] = cnt;
            tnc = MAXCHR;
          }
          tnc++;
        }
        cnt = override;
      }
    }
    else
    {
      cnt = get_free_character();
    }
    if (cnt != MAXCHR)
    {
      // IMPORTANT!!!
      chrindolist[cnt] = bfalse;
      chrisequipped[cnt] = bfalse;
      chrsparkle[cnt] = NOSPARKLE;
      chroverlay[cnt] = bfalse;
      chrmissilehandler[cnt] = cnt;

      // SetXY stuff...  Just in case
      tnc = 0;
      while (tnc < MAXSTOR)
      {
        chraix[cnt][tnc] = 0;
        chraiy[cnt][tnc] = 0;
        tnc++;
      }

      // Set up model stuff
      chron[cnt] = btrue;
      chrreloadtime[cnt] = 0;
      chrinwhichhand[cnt] = GRIPLEFT;
      //chrwaskilled[cnt] = bfalse;
      chrinpack[cnt] = bfalse;
      chrnextinpack[cnt] = MAXCHR;
      chrnuminpack[cnt] = 0;
      chrmodel[cnt] = profile;
      chrbasemodel[cnt] = profile;
      chrstoppedby[cnt] = capstoppedby[profile];
      chrlifeheal[cnt] = caplifeheal[profile];
      chrmanacost[cnt] = capmanacost[profile];
      chrinwater[cnt] = bfalse;
      chrnameknown[cnt] = capnameknown[profile];
      chrammoknown[cnt] = capnameknown[profile];
      chrhitready[cnt] = btrue;
      chrboretime[cnt] = BORETIME;
      chrcarefultime[cnt] = CAREFULTIME;
      chrcanbecrushed[cnt] = bfalse;
      chrdamageboost[cnt] = 0;
      chricon[cnt] = capicon[profile];


      // Enchant stuff
      chrfirstenchant[cnt] = MAXENCHANT;
      chrundoenchant[cnt] = MAXENCHANT;
      chrcanseeinvisible[cnt] = capcanseeinvisible[profile];
      chrcanchannel[cnt] = bfalse;
      chrmissiletreatment[cnt] = MISNORMAL;
      chrmissilecost[cnt] = 0;

	  //Skill Expansions
      chrcanseekurse[cnt] = capcanseekurse[profile];
      chrcanusedivine[cnt] = capcanusedivine[profile];
      chrcanusearcane[cnt] = capcanusearcane[profile];
      chrcandisarm[cnt] = capcandisarm[profile];
	  chrcanjoust[cnt] = capcanjoust[profile];
	  chrcanusetech[cnt] = capcanusetech[profile];
	  chrcanusepoison[cnt] = capcanusepoison[profile];
  	  chrcanuseadvancedweapons[cnt] = capcanuseadvancedweapons[profile];
	  chrcanbackstab[cnt] = capcanbackstab[profile];
	  chrcanread[cnt] = capcanread[profile];


      // Kurse state
      chriskursed[cnt] = ((rand() % 100) < capkursechance[profile]);
      if (!capisitem[profile])  chriskursed[cnt] = bfalse;


      // Ammo
      chrammomax[cnt] = capammomax[profile];
      chrammo[cnt] = capammo[profile];


      // Gender
      chrgender[cnt] = capgender[profile];
      if (chrgender[cnt] == GENRANDOM)  chrgender[cnt] = GENFEMALE + (rand() & 1);



      // Team stuff
      chrteam[cnt] = team;
      chrbaseteam[cnt] = team;
      chrcounter[cnt] = teammorale[team];
      if (!capinvictus[profile])  teammorale[team]++;
      chrorder[cnt] = 0;
      // Firstborn becomes the leader
      if (teamleader[team] == NOLEADER)
      {
        teamleader[team] = cnt;
      }
      // Skin
      if (capskinoverride[profile] != NOSKINOVERRIDE)
      {
        skin = capskinoverride[profile] % MAXSKIN;
      }
      if (skin >= madskins[profile])
      {
        skin = 0;
        if (madskins[profile] > 1)
        {
          skin = rand() % madskins[profile];
        }
      }
      chrtexture[cnt] = madskinstart[profile] + skin;
      // Life and Mana
      chralive[cnt] = btrue;
      chrlifecolor[cnt] = caplifecolor[profile];
      chrmanacolor[cnt] = capmanacolor[profile];
      chrlifemax[cnt] = generate_number(caplifebase[profile], capliferand[profile]);
      chrlife[cnt] = chrlifemax[cnt];
      chrlifereturn[cnt] = caplifereturn[profile];
      chrmanamax[cnt] = generate_number(capmanabase[profile], capmanarand[profile]);
      chrmanaflow[cnt] = generate_number(capmanaflowbase[profile], capmanaflowrand[profile]);
      chrmanareturn[cnt] = generate_number(capmanareturnbase[profile], capmanareturnrand[profile]); //>> MANARETURNSHIFT;
      chrmana[cnt] = chrmanamax[cnt];
      // SWID
      chrstrength[cnt] = generate_number(capstrengthbase[profile], capstrengthrand[profile]);
      chrwisdom[cnt] = generate_number(capwisdombase[profile], capwisdomrand[profile]);
      chrintelligence[cnt] = generate_number(capintelligencebase[profile], capintelligencerand[profile]);
      chrdexterity[cnt] = generate_number(capdexteritybase[profile], capdexterityrand[profile]);
      // Damage
      chrdefense[cnt] = capdefense[profile][skin];
      chrreaffirmdamagetype[cnt] = capattachedprtreaffirmdamagetype[profile];
      chrdamagetargettype[cnt] = capdamagetargettype[profile];
      tnc = 0;
      while (tnc < MAXDAMAGETYPE)
      {
        chrdamagemodifier[cnt][tnc] = capdamagemodifier[profile][tnc][skin];
        tnc++;
      }
      // AI stuff
      chraitype[cnt] = madai[chrmodel[cnt]];
      chrisplayer[cnt] = bfalse;
      chrislocalplayer[cnt] = bfalse;
      chralert[cnt] = ALERTIFSPAWNED;
      chraistate[cnt] = capstateoverride[profile];
      chraicontent[cnt] = capcontentoverride[profile];
      chraitarget[cnt] = cnt;
      chraiowner[cnt] = cnt;
      chraichild[cnt] = cnt;
      chraitime[cnt] = 0;
      chrlatchx[cnt] = 0;
      chrlatchy[cnt] = 0;
      chrlatchbutton[cnt] = 0;
      chrturnmode[cnt] = TURNMODEVELOCITY;
      // Flags
      chrstickybutt[cnt] = capstickybutt[profile];
      chropenstuff[cnt] = capcanopenstuff[profile];
      chrtransferblend[cnt] = captransferblend[profile];
      chrenviro[cnt] = capenviro[profile];
      chrwaterwalk[cnt] = capwaterwalk[profile];
      chrplatform[cnt] = capplatform[profile];
      chrisitem[cnt] = capisitem[profile];
      chrinvictus[cnt] = capinvictus[profile];
      chrismount[cnt] = capismount[profile];
      chrcangrabmoney[cnt] = capcangrabmoney[profile];
      // Jumping
      chrjump[cnt] = capjump[profile];
      chrjumpready[cnt] = btrue;
      chrjumpnumber[cnt] = 1;
      chrjumpnumberreset[cnt] = capjumpnumber[profile];
      chrjumptime[cnt] = JUMPDELAY;
      // Other junk
      chrflyheight[cnt] = capflyheight[profile];
      chrmaxaccel[cnt] = capmaxaccel[profile][skin];
      chralpha[cnt] = capalpha[profile];
      chrlight[cnt] = caplight[profile];
      chrflashand[cnt] = capflashand[profile];
      chrsheen[cnt] = capsheen[profile];
      chrdampen[cnt] = capdampen[profile];
      // Character size and bumping
      chrfat[cnt] = capsize[profile];
      chrsizegoto[cnt] = chrfat[cnt];
      chrsizegototime[cnt] = 0;
      chrshadowsize[cnt] = capshadowsize[profile] * chrfat[cnt];
      chrbumpsize[cnt] = capbumpsize[profile] * chrfat[cnt];
      chrbumpsizebig[cnt] = capbumpsizebig[profile] * chrfat[cnt];
      chrbumpheight[cnt] = capbumpheight[profile] * chrfat[cnt];

      chrshadowsizesave[cnt] = capshadowsize[profile];
      chrbumpsizesave[cnt] = capbumpsize[profile];
      chrbumpsizebigsave[cnt] = capbumpsizebig[profile];
      chrbumpheightsave[cnt] = capbumpheight[profile];

      chrbumpdampen[cnt] = capbumpdampen[profile];
      chrweight[cnt] = capweight[profile] * chrfat[cnt];
      if (capweight[profile] == 255) chrweight[cnt] = 65535;
      chrbumplast[cnt] = cnt;
      chrattacklast[cnt] = MAXCHR;
      chrhitlast[cnt] = cnt;
      // Grip info
      chrattachedto[cnt] = MAXCHR;
      chrholdingwhich[cnt][0] = MAXCHR;
      chrholdingwhich[cnt][1] = MAXCHR;
      // Image rendering
      chruoffset[cnt] = 0;
      chrvoffset[cnt] = 0;
      chruoffvel[cnt] = capuoffvel[profile];
      chrvoffvel[cnt] = capvoffvel[profile];
      chrredshift[cnt] = 0;
      chrgrnshift[cnt] = 0;
      chrblushift[cnt] = 0;
      // Movement
      chrsneakspd[cnt] = capsneakspd[profile];
      chrwalkspd[cnt] = capwalkspd[profile];
      chrrunspd[cnt] = caprunspd[profile];


      // Set up position
      chrxpos[cnt] = x;
      chrypos[cnt] = y;
      chroldx[cnt] = x;
      chroldy[cnt] = y;
      chrturnleftright[cnt] = facing;
      chrlightturnleftright[cnt] = 0;
      chronwhichfan[cnt] = mesh_get_fan(chrxpos[cnt], chrypos[cnt]);
      chrlevel[cnt] = get_level(chrxpos[cnt], chrypos[cnt], chronwhichfan[cnt], chrwaterwalk[cnt]) + RAISE;
      if (z < chrlevel[cnt]) z = chrlevel[cnt];
      chrzpos[cnt] = z;
      chroldz[cnt] = z;
      chrxstt[cnt] = chrxpos[cnt];
      chrystt[cnt] = chrypos[cnt];
      chrzstt[cnt] = chrzpos[cnt];
      chrxvel[cnt] = 0;
      chryvel[cnt] = 0;
      chrzvel[cnt] = 0;
      chrxtrgvel[cnt] = 0;
      chrytrgvel[cnt] = 0;
      chrztrgvel[cnt] = 0;
      chrturnmaplr[cnt] = 32768;  // These two mean on level surface
      chrturnmapud[cnt] = 32768;
      chrscale[cnt] = chrfat[cnt] * madscale[chrmodel[cnt]] * 4;


      // AI and action stuff
      chraigoto[cnt] = 0;
      chraigotoadd[cnt] = 1;
      chraigotox[cnt][0] = chrxpos[cnt];
      chraigotoy[cnt][0] = chrypos[cnt];
      chractionready[cnt] = btrue;
      chrkeepaction[cnt] = bfalse;
      chrloopaction[cnt] = bfalse;
      chraction[cnt] = ACTIONDA;
      chrnextaction[cnt] = ACTIONDA;
      chrlip[cnt] = 0;
      chrflip[cnt] = 0.0f;
      chrframe[cnt] = madframestart[chrmodel[cnt]];
      chrlastframe[cnt] = chrframe[cnt];
      chrpassage[cnt] = 0;
      chrholdingweight[cnt] = 0;
      chronwhichplatform[cnt] = MAXCHR;


      // Timers set to 0
      chrgrogtime[cnt] = 0;
      chrdazetime[cnt] = 0;


      // Money is added later
      chrmoney[cnt] = capmoney[profile];


      // Name the character
      if (name == NULL)
      {
        // Generate a random name
        naming_names(profile);
        strncpy(chrname[cnt], namingnames, sizeof(chrname[cnt]));
      }
      else
      {
        // A name has been given
        tnc = 0;
        while (tnc < MAXCAPNAMESIZE - 1)
        {
          chrname[cnt][tnc] = name[tnc];
          tnc++;
        }
        chrname[cnt][tnc] = 0;
      }

      // Set up initial fade in lighting
      tnc = 0;
      while (tnc < madtransvertices[chrmodel[cnt]])
      {
        chrvrta[cnt][tnc] = 0;
        tnc++;
      }


      // Particle attachments
      tnc = 0;
      while (tnc < capattachedprtamount[profile])
      {
        spawn_one_particle(chrxpos[cnt], chrypos[cnt], chrzpos[cnt],
                           0, chrmodel[cnt], capattachedprttype[profile],
                           cnt, SPAWNLAST + tnc, chrteam[cnt], cnt, tnc, MAXCHR);
        tnc++;
      }
      chrreaffirmdamagetype[cnt] = capattachedprtreaffirmdamagetype[profile];


      // Experience
      tnc = generate_number(capexperiencebase[profile], capexperiencerand[profile]);
      if (tnc > MAXXP) tnc = MAXXP;
      chrexperience[cnt] = tnc;
      chrexperiencelevel[cnt] = capleveloverride[profile];
    }
  }
  return cnt;
}

//--------------------------------------------------------------------------------------------
void respawn_character(Uint16 character)
{
  // ZZ> This function respawns a character
  Uint16 item;

  if (!chralive[character])
  {
    spawn_poof(character, chrmodel[character]);
    disaffirm_attached_particles(character);
    chralive[character] = btrue;
    chrboretime[character] = BORETIME;
    chrcarefultime[character] = CAREFULTIME;
    chrlife[character] = chrlifemax[character];
    chrmana[character] = chrmanamax[character];
    chrxpos[character] = chrxstt[character];
    chrypos[character] = chrystt[character];
    chrzpos[character] = chrzstt[character];
    chrxvel[character] = 0;
    chryvel[character] = 0;
    chrzvel[character] = 0;
    chrxtrgvel[character] = 0;
    chrytrgvel[character] = 0;
    chrztrgvel[character] = 0;
    chrteam[character] = chrbaseteam[character];
    chrcanbecrushed[character] = bfalse;
    chrturnmaplr[character] = 32768;  // These two mean on level surface
    chrturnmapud[character] = 32768;
    if (teamleader[chrteam[character]] == NOLEADER)  teamleader[chrteam[character]] = character;
    if (!chrinvictus[character])  teammorale[chrbaseteam[character]]++;
    chractionready[character] = btrue;
    chrkeepaction[character] = bfalse;
    chrloopaction[character] = bfalse;
    chraction[character] = ACTIONDA;
    chrnextaction[character] = ACTIONDA;
    chrlip[character] = 0;
    chrflip[character] = 0.0f;
    chrframe[character] = madframestart[chrmodel[character]];
    chrlastframe[character] = chrframe[character];
    chrplatform[character] = capplatform[chrmodel[character]];
    chrflyheight[character] = capflyheight[chrmodel[character]];
    chrbumpdampen[character] = capbumpdampen[chrmodel[character]];
    chrbumpsize[character] = capbumpsize[chrmodel[character]] * chrfat[character];
    chrbumpsizebig[character] = capbumpsizebig[chrmodel[character]] * chrfat[character];
    chrbumpheight[character] = capbumpheight[chrmodel[character]] * chrfat[character];

    chrbumpsizesave[character] = capbumpsize[chrmodel[character]];
    chrbumpsizebigsave[character] = capbumpsizebig[chrmodel[character]];
    chrbumpheightsave[character] = capbumpheight[chrmodel[character]];

    //        chralert[character] = ALERTIFSPAWNED;
    chralert[character] = 0;
    //        chraistate[character] = 0;
    chraitarget[character] = character;
    chraitime[character] = 0;
    chrgrogtime[character] = 0;
    chrdazetime[character] = 0;
    reaffirm_attached_particles(character);


    // Let worn items come back
    item = chrnextinpack[character];
    while (item != MAXCHR)
    {
      if (chrisequipped[item])
      {
        chrisequipped[item] = bfalse;
        chralert[item] |= ALERTIFATLASTWAYPOINT;  // doubles as PutAway
      }
      item = chrnextinpack[item];
    }
  }
}

//--------------------------------------------------------------------------------------------
Uint16 change_armor(Uint16 character, Uint16 skin)
{
  // ZZ> This function changes the armor of the character
  Uint16 enchant, sTmp;
  int iTmp, cnt;


  // Remove armor enchantments
  enchant = chrfirstenchant[character];
  while (enchant < MAXENCHANT)
  {
    for(cnt=SETSLASHMODIFIER; cnt<=SETZAPMODIFIER; cnt++)
    {
      unset_enchant_value(enchant, cnt);
    }
    enchant = encnextenchant[enchant];
  }


  // Change the skin
  sTmp = chrmodel[character];
  if (skin > madskins[sTmp])  skin = 0;
  chrtexture[character] = madskinstart[sTmp] + skin;


  // Change stats associated with skin
  chrdefense[character] = capdefense[sTmp][skin];
  iTmp = 0;
  while (iTmp < MAXDAMAGETYPE)
  {
    chrdamagemodifier[character][iTmp] = capdamagemodifier[sTmp][iTmp][skin];
    iTmp++;
  }
  chrmaxaccel[character] = capmaxaccel[sTmp][skin];


  // Reset armor enchantments
  // These should really be done in reverse order ( Start with last enchant ), but
  // I don't care at this point !!!BAD!!!
  enchant = chrfirstenchant[character];
  while (enchant < MAXENCHANT)
  {
    for(cnt=SETSLASHMODIFIER; cnt<=SETZAPMODIFIER; cnt++)
    {
      set_enchant_value(enchant, cnt, enceve[enchant]);
    };
    add_enchant_value(enchant, ADDACCEL, enceve[enchant]);
    add_enchant_value(enchant, ADDDEFENSE, enceve[enchant]);
    enchant = encnextenchant[enchant];
  }
  return skin;
}

//--------------------------------------------------------------------------------------------
void change_character(Uint16 cnt, Uint16 profile, Uint8 skin,
                      Uint8 leavewhich)
{
  // ZZ> This function polymorphs a character, changing stats, dropping weapons
  int tnc, enchant;
  Uint16 sTmp, item;

  if(profile > MAXMODEL || !madused[profile]) return;


  // Drop left weapon
  sTmp = chrholdingwhich[cnt][0];
  if (sTmp != MAXCHR && (!capgripvalid[profile][0] || capismount[profile]))
  {
    detach_character_from_mount(sTmp, btrue, btrue);
    if (chrismount[cnt])
    {
      chrzvel[sTmp] = DISMOUNTZVEL;
      chrzpos[sTmp] += DISMOUNTZVEL;
      chrjumptime[sTmp] = JUMPDELAY;
    }
  }


  // Drop right weapon
  sTmp = chrholdingwhich[cnt][1];
  if (sTmp != MAXCHR && !capgripvalid[profile][1])
  {
    detach_character_from_mount(sTmp, btrue, btrue);
    if (chrismount[cnt])
    {
      chrzvel[sTmp] = DISMOUNTZVEL;
      chrzpos[sTmp] += DISMOUNTZVEL;
      chrjumptime[sTmp] = JUMPDELAY;
    }
  }


  // Remove particles
  disaffirm_attached_particles(cnt);


  // Remove enchantments
  if (leavewhich == LEAVEFIRST)
  {
    // Remove all enchantments except top one
    enchant = chrfirstenchant[cnt];
    if (enchant != MAXENCHANT)
    {
      while (encnextenchant[enchant] != MAXENCHANT)
      {
        remove_enchant(encnextenchant[enchant]);
      }
    }
  }
  if (leavewhich == LEAVENONE)
  {
    // Remove all enchantments
    disenchant_character(cnt);
  }


  // Stuff that must be set
  chrmodel[cnt] = profile;
  chrstoppedby[cnt] = capstoppedby[profile];
  chrlifeheal[cnt] = caplifeheal[profile];
  chrmanacost[cnt] = capmanacost[profile];
  // Ammo
  chrammomax[cnt] = capammomax[profile];
  chrammo[cnt] = capammo[profile];
  // Gender
  if (capgender[profile] != GENRANDOM) // GENRANDOM means keep old gender
  {
    chrgender[cnt] = capgender[profile];
  }


  // AI stuff
  chraitype[cnt] = madai[profile];
  chraistate[cnt] = 0;
  chraitime[cnt] = 0;
  chrlatchx[cnt] = 0;
  chrlatchy[cnt] = 0;
  chrlatchbutton[cnt] = 0;
  chrturnmode[cnt] = TURNMODEVELOCITY;
  // Flags
  chrstickybutt[cnt] = capstickybutt[profile];
  chropenstuff[cnt] = capcanopenstuff[profile];
  chrtransferblend[cnt] = captransferblend[profile];
  chrenviro[cnt] = capenviro[profile];
  chrplatform[cnt] = capplatform[profile];
  chrisitem[cnt] = capisitem[profile];
  chrinvictus[cnt] = capinvictus[profile];
  chrismount[cnt] = capismount[profile];
  chrcangrabmoney[cnt] = capcangrabmoney[profile];
  chrjumptime[cnt] = JUMPDELAY;
  // Character size and bumping
  chrshadowsize[cnt] = capshadowsize[profile] * chrfat[cnt];
  chrbumpsize[cnt] = capbumpsize[profile] * chrfat[cnt];
  chrbumpsizebig[cnt] = capbumpsizebig[profile] * chrfat[cnt];
  chrbumpheight[cnt] = capbumpheight[profile] * chrfat[cnt];

  chrshadowsizesave[cnt] = capshadowsize[profile];
  chrbumpsizesave[cnt] = capbumpsize[profile];
  chrbumpsizebigsave[cnt] = capbumpsizebig[profile];
  chrbumpheightsave[cnt] = capbumpheight[profile];

  chrbumpdampen[cnt] = capbumpdampen[profile];
  chrweight[cnt] = capweight[profile] * chrfat[cnt];
  if (capweight[profile] == 255) chrweight[cnt] = 65535;
  // Character scales...  Magic numbers
  if (chrattachedto[cnt] == MAXCHR)
  {
    chrscale[cnt] = chrfat[cnt] * madscale[profile] * 4;
  }
  else
  {
    chrscale[cnt] = chrfat[cnt] / (chrfat[chrattachedto[cnt]] * 1280);
    tnc = madvertices[chrmodel[chrattachedto[cnt]]] - chrinwhichhand[cnt];
    chrweapongrip[cnt][0] = tnc;
    chrweapongrip[cnt][1] = tnc + 1;
    chrweapongrip[cnt][2] = tnc + 2;
    chrweapongrip[cnt][3] = tnc + 3;
  }

  item = chrholdingwhich[cnt][0];
  if (item != MAXCHR)
  {
    chrscale[item] = chrfat[item] / (chrfat[cnt] * 1280);
    tnc = madvertices[chrmodel[cnt]] - GRIPLEFT;
    chrweapongrip[item][0] = tnc;
    chrweapongrip[item][1] = tnc + 1;
    chrweapongrip[item][2] = tnc + 2;
    chrweapongrip[item][3] = tnc + 3;
  }

  item = chrholdingwhich[cnt][1];
  if (item != MAXCHR)
  {
    chrscale[item] = chrfat[item] / (chrfat[cnt] * 1280);
    tnc = madvertices[chrmodel[cnt]] - GRIPRIGHT;
    chrweapongrip[item][0] = tnc;
    chrweapongrip[item][1] = tnc + 1;
    chrweapongrip[item][2] = tnc + 2;
    chrweapongrip[item][3] = tnc + 3;
  }
  // Image rendering
  chruoffset[cnt] = 0;
  chrvoffset[cnt] = 0;
  chruoffvel[cnt] = capuoffvel[profile];
  chrvoffvel[cnt] = capvoffvel[profile];
  // Movement
  chrsneakspd[cnt] = capsneakspd[profile];
  chrwalkspd[cnt] = capwalkspd[profile];
  chrrunspd[cnt] = caprunspd[profile];


  // AI and action stuff
  chractionready[cnt] = btrue;
  chrkeepaction[cnt] = bfalse;
  chrloopaction[cnt] = bfalse;
  chraction[cnt] = ACTIONDA;
  chrnextaction[cnt] = ACTIONDA;
  chrlip[cnt] = 0;
  chrflip[cnt] = 0.0f;
  chrframe[cnt] = madframestart[profile];
  chrlastframe[cnt] = chrframe[cnt];
  chrholdingweight[cnt] = 0;
  chronwhichplatform[cnt] = MAXCHR;


  // Set the skin
  change_armor(cnt, skin);


  // Reaffirm them particles...
  chrreaffirmdamagetype[cnt] = capattachedprtreaffirmdamagetype[profile];
  reaffirm_attached_particles(cnt);


  // Set up initial fade in lighting
  tnc = 0;
  while (tnc < madtransvertices[chrmodel[cnt]])
  {
    chrvrta[cnt][tnc] = 0;
    tnc++;
  }
}

//--------------------------------------------------------------------------------------------
Uint16 chr_get_target_in_block(int block_x, int block_y, Uint16 character, bool_t items,
                           bool_t friends, bool_t enemies, bool_t dead, bool_t seeinvisible, IDSZ idsz,
                           bool_t excludeid)
{
  // ZZ> This is a good little helper, that returns != MAXCHR if a suitable target
  //     was found

  int cnt;
  Uint16 charb;
  Uint32 fanblock;
  Uint8 team;

  if(MAXCHR==character || !chron[character] || chrinpack[character]) return MAXCHR; 
  if(block_x<0 || block_x>(meshsizex>>2) || block_y<0 || block_y>(meshsizey>>2)) return MAXCHR;

  fanblock = block_x + meshblockstart[block_y];
  team = chrteam[character];
  charb = meshbumplistchr[fanblock];
  for (cnt=0; cnt < meshbumplistchrnum[fanblock]; cnt++, charb = chrbumpnext[charb])
  {
    if(character == charb || chrattachedto[character] == charb || chrattachedto[charb]==character) continue;

    if (dead != chralive[charb] && (seeinvisible || (chralpha[charb] + chrlight[charb] > INVISIBLE)))
    {
      if ((enemies && teamhatesteam[team][chrteam[charb]] && !chrinvictus[charb]) ||
          (items && chrisitem[charb]) ||
          (friends && chrbaseteam[charb] == team))
      {
        if (chrattachedto[character] != charb)
        {
          if (!chrisitem[charb] || items)
          {
            if (idsz != IDSZNONE)
            {
              if (capidsz[chrmodel[charb]][IDSZPARENT] == idsz ||
                  capidsz[chrmodel[charb]][IDSZTYPE] == idsz)
              {
                if (!excludeid) return charb;
              }
              else
              {
                if (excludeid)  return charb;
              }
            }
            else
            {
              return charb;
            }
          }
        }
      }
    }
  }

  return MAXCHR;
}

//--------------------------------------------------------------------------------------------
Uint16 chr_get_nearby_target(Uint16 character, bool_t items,
                                 bool_t friends, bool_t enemies, bool_t dead, IDSZ idsz)
{
  // ZZ> This function finds a nearby target, or it returns MAXCHR if it can't find one
  int x, y;
  bool_t seeinvisible;
  Uint16 found;
  seeinvisible = chrcanseeinvisible[character];

  if(MAXCHR==character || !chron[character] || chrinpack[character]) return MAXCHR;

  // Current fanblock
  x = ((int)chrxpos[character]) >> 9;
  y = ((int)chrypos[character]) >> 9;
  found = chr_get_target_in_block(x, y, character, items, friends, enemies, dead, seeinvisible, idsz, 0);

  return (found==character) ? MAXCHR : found;
}

//--------------------------------------------------------------------------------------------
bool_t cost_mana(Uint16 character, int amount, Uint16 killer)
{
  // ZZ> This function takes mana from a character ( or gives mana ),
  //     and returns btrue if the character had enough to pay, or bfalse
  //     otherwise
  int iTmp;


  iTmp = chrmana[character] - amount;
  if (iTmp < 0)
  {
    chrmana[character] = 0;
    if (chrcanchannel[character])
    {
      chrlife[character] += iTmp;
      if (chrlife[character] <= 0)
      {
        kill_character(character, character);
      }
      return btrue;
    }
    return bfalse;
  }
  else
  {
    chrmana[character] = iTmp;
    if (iTmp > chrmanamax[character])
    {
      chrmana[character] = chrmanamax[character];
    }
  }
  return btrue;
}

//--------------------------------------------------------------------------------------------
Uint16 chr_find_distant_target(Uint16 character, int maxdistance, bool_t enemies)
{
  // ZZ> This function finds a target, or it returns MAXCHR if it can't find one...
  //     maxdistance should be the square of the actual distance you want to use
  //     as the cutoff...
  int cnt, distance, xdistance, ydistance;
  Uint8 team;

  if(MAXCHR==character || !chron[character] || chrinpack[character]) return MAXCHR;

  team = chrteam[character];
  for (cnt = 0; cnt < MAXCHR; cnt++)
  {
    if(character == cnt || chrattachedto[character] == cnt || chrattachedto[cnt]==character) continue;
    if (!chron[cnt] || !chralive[cnt] || chrinpack[cnt]) continue;
    if(enemies && (!teamhatesteam[team][chrteam[cnt]] || chrinvictus[cnt]) ) continue;


    if (chrcanseeinvisible[character] || (chralpha[cnt] + chrlight[cnt] > INVISIBLE))
    {
      xdistance = chrxpos[cnt] - chrxpos[character];
      ydistance = chrypos[cnt] - chrypos[character];
      distance = xdistance * xdistance + ydistance * ydistance;
      if (distance < maxdistance)
      {
        return cnt;
      }
    }
  }

  return MAXCHR;
}

//--------------------------------------------------------------------------------------------
void switch_team(Uint16 character, Uint8 team)
{
  // ZZ> This function makes a character join another team...
  if (team < MAXTEAM)
  {
    if (!chrinvictus[character])
    {
      teammorale[chrbaseteam[character]]--;
      teammorale[team]++;
    }
    if ((!chrismount[character] || chrholdingwhich[character][0] == MAXCHR) &&
        (!chrisitem[character] || chrattachedto[character] == MAXCHR))
    {
      chrteam[character] = team;
    }
    chrbaseteam[character] = team;
    if (teamleader[team] == NOLEADER)
    {
      teamleader[team] = character;
    }
  }
}

//--------------------------------------------------------------------------------------------
void chr_get_nearest_in_block(int block_x, int block_y, Uint16 character, bool_t items,
                              bool_t friends, bool_t enemies, bool_t dead, bool_t seeinvisible, IDSZ idsz)
{
  // ZZ> This is a good little helper
  float distance, xdis, ydis;
  int cnt;
  Uint8 team;
  Uint16 charb;
  Uint32 fanblock;

  if(MAXCHR==character || !chron[character] || chrinpack[character]) return;

  if(block_x<0 || block_x>(meshsizex>>2) || block_y<0 || block_y>(meshsizey>>2)) return;

  fanblock = block_x + meshblockstart[block_y];
  team     = chrteam[character];
  charb    = meshbumplistchr[fanblock];
  for (cnt = 0; cnt < meshbumplistchrnum[fanblock]; cnt++, charb = chrbumpnext[charb])
  {
    if(character == charb || chrattachedto[character] == charb || chrattachedto[charb]==character) continue;

    if (dead != chralive[charb] && (seeinvisible || (chralpha[charb] + chrlight[charb] > INVISIBLE)))
    {
      if ((enemies && teamhatesteam[team][chrteam[charb]]) ||
          (items && chrisitem[charb]) ||
          (friends && chrteam[charb] == team) ||
          (friends && enemies))
      {
        if (charb != character && chrattachedto[character] != charb && chrattachedto[charb] == MAXCHR && !chrinpack[charb])
        {
          if (!chrinvictus[charb] || items)
          {
            if (idsz != IDSZNONE)
            {
              if (capidsz[chrmodel[charb]][IDSZPARENT] == idsz ||
                  capidsz[chrmodel[charb]][IDSZTYPE] == idsz)
              {
                xdis = chrxpos[character] - chrxpos[charb];
                ydis = chrypos[character] - chrypos[charb];
                xdis *= xdis;
                ydis *= ydis;
                distance = xdis + ydis;
                if (distance < globaldistance)
                {
                  globalnearest = charb;
                  globaldistance = distance;
                }
              }
            }
            else
            {
              xdis = chrxpos[character] - chrxpos[charb];
              ydis = chrypos[character] - chrypos[charb];
              xdis *= xdis;
              ydis *= ydis;
              distance = xdis + ydis;
              if (distance < globaldistance)
              {
                globalnearest = charb;
                globaldistance = distance;
              }
            }
          }
        }
      }
    }
  }

  return;
}

//--------------------------------------------------------------------------------------------
Uint16 chr_get_nearest_target(Uint16 character, bool_t items,
                                  bool_t friends, bool_t enemies, bool_t dead, IDSZ idsz)
{
  // ZZ> This function finds an target, or it returns MAXCHR if it can't find one
  int x, y;
  char seeinvisible;
  seeinvisible = chrcanseeinvisible[character];

  if(MAXCHR==character || !chron[character] || chrinpack[character]) return MAXCHR;


  // Current fanblock
  x = ((int)chrxpos[character]) >> 9;
  y = ((int)chrypos[character]) >> 9;

  globalnearest = MAXCHR;
  globaldistance = 999999;
  chr_get_nearest_in_block(x + 0, y + 0, character, items, friends, enemies, dead, seeinvisible, idsz);

  chr_get_nearest_in_block(x - 1, y + 0, character, items, friends, enemies, dead, seeinvisible, idsz);
  chr_get_nearest_in_block(x + 1, y + 0, character, items, friends, enemies, dead, seeinvisible, idsz);
  chr_get_nearest_in_block(x + 0, y - 1, character, items, friends, enemies, dead, seeinvisible, idsz);
  chr_get_nearest_in_block(x + 0, y + 1, character, items, friends, enemies, dead, seeinvisible, idsz);

  chr_get_nearest_in_block(x - 1, y + 1, character, items, friends, enemies, dead, seeinvisible, idsz);
  chr_get_nearest_in_block(x + 1, y - 1, character, items, friends, enemies, dead, seeinvisible, idsz);
  chr_get_nearest_in_block(x - 1, y - 1, character, items, friends, enemies, dead, seeinvisible, idsz);
  chr_get_nearest_in_block(x + 1, y + 1, character, items, friends, enemies, dead, seeinvisible, idsz);

  if(globalnearest == character)
    return MAXCHR;
  else
    return globalnearest;
}

//--------------------------------------------------------------------------------------------
Uint16 chr_get_wide_target(Uint16 character, bool_t items,
                               bool_t friends, bool_t enemies, bool_t dead, Uint32 idsz, bool_t excludeid)
{
  // ZZ> This function finds an target, or it returns MAXCHR if it can't find one
  int x, y;
  Uint16 enemy;
  char seeinvisible;
  seeinvisible = chrcanseeinvisible[character];

  if(MAXCHR==character || !chron[character] || chrinpack[character]) return MAXCHR;

  // Current fanblock
  x = ((int)chrxpos[character]) >> 9;
  y = ((int)chrypos[character]) >> 9;

  enemy = chr_get_target_in_block(x + 0, y + 0, character, items, friends, enemies, dead, seeinvisible, idsz, excludeid);
  if (enemy != MAXCHR && enemy != character)  return enemy;

  enemy = chr_get_target_in_block(x - 1, y + 0, character, items, friends, enemies, dead, seeinvisible, idsz, excludeid);
  if (enemy != MAXCHR && enemy != character)  return enemy;
  enemy = chr_get_target_in_block(x + 1, y + 0, character, items, friends, enemies, dead, seeinvisible, idsz, excludeid);
  if (enemy != MAXCHR && enemy != character)  return enemy;
  enemy = chr_get_target_in_block(x + 0, y - 1, character, items, friends, enemies, dead, seeinvisible, idsz, excludeid);
  if (enemy != MAXCHR && enemy != character)  return enemy;
  enemy = chr_get_target_in_block(x + 0, y + 1, character, items, friends, enemies, dead, seeinvisible, idsz, excludeid);
  if (enemy != MAXCHR && enemy != character)  return enemy;

  enemy = chr_get_target_in_block(x - 1, y + 1, character, items, friends, enemies, dead, seeinvisible, idsz, excludeid);
  if (enemy != MAXCHR && enemy != character)  return enemy;
  enemy = chr_get_target_in_block(x + 1, y - 1, character, items, friends, enemies, dead, seeinvisible, idsz, excludeid);
  if (enemy != MAXCHR && enemy != character)  return enemy;
  enemy = chr_get_target_in_block(x - 1, y - 1, character, items, friends, enemies, dead, seeinvisible, idsz, excludeid);
  if (enemy != MAXCHR && enemy != character)  return enemy;
  enemy = chr_get_target_in_block(x + 1, y + 1, character, items, friends, enemies, dead, seeinvisible, idsz, excludeid);
  if (enemy != MAXCHR && enemy != character)  return enemy;

  return MAXCHR;
}

//--------------------------------------------------------------------------------------------
void issue_clean(Uint16 character)
{
  // ZZ> This function issues a clean up order to all teammates
  Uint8 team;
  Uint16 cnt;


  team = chrteam[character];
  cnt = 0;
  while (cnt < MAXCHR)
  {
    if (chrteam[cnt] == team && !chralive[cnt])
    {
      chraitime[cnt] = 2;  // Don't let it think too much...
      chralert[cnt] = ALERTIFCLEANEDUP;
    }
    cnt++;
  }
}

//--------------------------------------------------------------------------------------------
int restock_ammo(Uint16 character, IDSZ idsz)
{
  // ZZ> This function restocks the characters ammo, if it needs ammo and if
  //     either its parent or type idsz match the given idsz.  This
  //     function returns the amount of ammo given.
  int amount, model;

  amount = 0;
  if (character < MAXCHR)
  {
    if (chron[character])
    {
      model = chrmodel[character];
      if (capidsz[model][IDSZPARENT] == idsz || capidsz[model][IDSZTYPE] == idsz)
      {
        if (chrammo[character] < chrammomax[character])
        {
          amount = chrammomax[character] - chrammo[character];
          chrammo[character] = chrammomax[character];
        }
      }
    }
  }
  return amount;
}

//--------------------------------------------------------------------------------------------
void issue_order(Uint16 character, Uint32 order)
{
  // ZZ> This function issues an order for help to all teammates
  Uint8 team;
  Uint8 counter;
  Uint16 cnt;


  team = chrteam[character];
  counter = 0;
  cnt = 0;
  while (cnt < MAXCHR)
  {
    if (chrteam[cnt] == team)
    {
      chrorder[cnt] = order;
      chrcounter[cnt] = counter;
      chralert[cnt] |= ALERTIFORDERED;
      counter++;
    }
    cnt++;
  }
}

//--------------------------------------------------------------------------------------------
void issue_special_order(Uint32 order, IDSZ idsz)
{
  // ZZ> This function issues an order to all characters with the a matching special IDSZ
  Uint8 counter;
  Uint16 cnt;


  counter = 0;
  cnt = 0;
  while (cnt < MAXCHR)
  {
    if (chron[cnt])
    {
      if (capidsz[chrmodel[cnt]][IDSZSPECIAL] == idsz)
      {
        chrorder[cnt] = order;
        chrcounter[cnt] = counter;
        chralert[cnt] |= ALERTIFORDERED;
        counter++;
      }
    }
    cnt++;
  }
}

//--------------------------------------------------------------------------------------------
void set_alerts(Uint16 character)
{
  // ZZ> This function polls some alert conditions
  if (chraitime[character] != 0)
  {
    chraitime[character]--;
  }
  if (chrxpos[character] < chraigotox[character][chraigoto[character]] + WAYTHRESH &&
      chrxpos[character] > chraigotox[character][chraigoto[character]] - WAYTHRESH &&
      chrypos[character] < chraigotoy[character][chraigoto[character]] + WAYTHRESH &&
      chrypos[character] > chraigotoy[character][chraigoto[character]] - WAYTHRESH)
  {
    chralert[character] |= ALERTIFATWAYPOINT;
    chraigoto[character]++;
    if (chraigoto[character] == chraigotoadd[character])
    {
      chraigoto[character] = 0;
      if (!capisequipment[chrmodel[character]])
      {
        chralert[character] |= ALERTIFATLASTWAYPOINT;
      }
    }
  }
}

//--------------------------------------------------------------------------------------------
void free_all_enchants()
{
  // ZZ> This functions frees all of the enchantments
  numfreeenchant = 0;
  while (numfreeenchant < MAXENCHANT)
  {
    freeenchant[numfreeenchant] = numfreeenchant;
    encon[numfreeenchant] = bfalse;
    numfreeenchant++;
  }
}

//--------------------------------------------------------------------------------------------
void load_one_enchant_type(char* szLoadName, Uint16 profile)
{
  // ZZ> This function loads the enchantment associated with an object
  FILE* fileread;
  char cTmp;
  int iTmp, tTmp;
  float fTmp;
  int num;
  IDSZ idsz;

  globalname = szLoadName;
  evevalid[profile] = bfalse;
  fileread = fs_fileOpen(PRI_NONE, NULL, szLoadName, "r");
  if (NULL != fileread)
  {
    evevalid[profile] = btrue;


    // btrue/bfalse values
    goto_colon(fileread);  cTmp = get_first_letter(fileread);
    everetarget[profile] = bfalse;
    if (cTmp == 'T' || cTmp == 't')  everetarget[profile] = btrue;
    goto_colon(fileread);  cTmp = get_first_letter(fileread);
    eveoverride[profile] = bfalse;
    if (cTmp == 'T' || cTmp == 't')  eveoverride[profile] = btrue;
    goto_colon(fileread);  cTmp = get_first_letter(fileread);
    everemoveoverridden[profile] = bfalse;
    if (cTmp == 'T' || cTmp == 't')  everemoveoverridden[profile] = btrue;
    goto_colon(fileread);  cTmp = get_first_letter(fileread);
    evekillonend[profile] = bfalse;
    if (cTmp == 'T' || cTmp == 't')  evekillonend[profile] = btrue;
    goto_colon(fileread);  cTmp = get_first_letter(fileread);
    evepoofonend[profile] = bfalse;
    if (cTmp == 'T' || cTmp == 't')  evepoofonend[profile] = btrue;


    // More stuff
    goto_colon(fileread);  fscanf(fileread, "%d", &iTmp);  evetime[profile] = iTmp;
    goto_colon(fileread);  fscanf(fileread, "%d", &iTmp);  eveendmessage[profile] = iTmp;


    // Drain stuff
    goto_colon(fileread);  fscanf(fileread, "%f", &fTmp);  eveownermana[profile] = fTmp * 256;
    goto_colon(fileread);  fscanf(fileread, "%f", &fTmp);  evetargetmana[profile] = fTmp * 256;
    goto_colon(fileread);  cTmp = get_first_letter(fileread);
    eveendifcantpay[profile] = bfalse;
    if (cTmp == 'T' || cTmp == 't')  eveendifcantpay[profile] = btrue;
    goto_colon(fileread);  fscanf(fileread, "%f", &fTmp);  eveownerlife[profile] = fTmp * 256;
    goto_colon(fileread);  fscanf(fileread, "%f", &fTmp);  evetargetlife[profile] = fTmp * 256;


    // Specifics
    goto_colon(fileread);  cTmp = get_first_letter(fileread);
    evedontdamagetype[profile] = DAMAGENULL;
    if (cTmp == 'S' || cTmp == 's')  evedontdamagetype[profile] = DAMAGESLASH;
    if (cTmp == 'C' || cTmp == 'c')  evedontdamagetype[profile] = DAMAGECRUSH;
    if (cTmp == 'P' || cTmp == 'p')  evedontdamagetype[profile] = DAMAGEPOKE;
    if (cTmp == 'H' || cTmp == 'h')  evedontdamagetype[profile] = DAMAGEHOLY;
    if (cTmp == 'E' || cTmp == 'e')  evedontdamagetype[profile] = DAMAGEEVIL;
    if (cTmp == 'F' || cTmp == 'f')  evedontdamagetype[profile] = DAMAGEFIRE;
    if (cTmp == 'I' || cTmp == 'i')  evedontdamagetype[profile] = DAMAGEICE;
    if (cTmp == 'Z' || cTmp == 'z')  evedontdamagetype[profile] = DAMAGEZAP;
    goto_colon(fileread);  cTmp = get_first_letter(fileread);
    eveonlydamagetype[profile] = DAMAGENULL;
    if (cTmp == 'S' || cTmp == 's')  eveonlydamagetype[profile] = DAMAGESLASH;
    if (cTmp == 'C' || cTmp == 'c')  eveonlydamagetype[profile] = DAMAGECRUSH;
    if (cTmp == 'P' || cTmp == 'p')  eveonlydamagetype[profile] = DAMAGEPOKE;
    if (cTmp == 'H' || cTmp == 'h')  eveonlydamagetype[profile] = DAMAGEHOLY;
    if (cTmp == 'E' || cTmp == 'e')  eveonlydamagetype[profile] = DAMAGEEVIL;
    if (cTmp == 'F' || cTmp == 'f')  eveonlydamagetype[profile] = DAMAGEFIRE;
    if (cTmp == 'I' || cTmp == 'i')  eveonlydamagetype[profile] = DAMAGEICE;
    if (cTmp == 'Z' || cTmp == 'z')  eveonlydamagetype[profile] = DAMAGEZAP;
    goto_colon(fileread);  everemovedbyidsz[profile] = get_idsz(fileread);


    // Now the set values
    num = 0;
    goto_colon(fileread);  cTmp = get_first_letter(fileread);
    evesetyesno[profile][num] = (cTmp == 'T' || cTmp == 't');
    cTmp = get_first_letter(fileread);
    evesetvalue[profile][num] = DAMAGESLASH;
    if (cTmp == 'C' || cTmp == 'c')  evesetvalue[profile][num] = DAMAGECRUSH;
    if (cTmp == 'P' || cTmp == 'p')  evesetvalue[profile][num] = DAMAGEPOKE;
    if (cTmp == 'H' || cTmp == 'h')  evesetvalue[profile][num] = DAMAGEHOLY;
    if (cTmp == 'E' || cTmp == 'e')  evesetvalue[profile][num] = DAMAGEEVIL;
    if (cTmp == 'F' || cTmp == 'f')  evesetvalue[profile][num] = DAMAGEFIRE;
    if (cTmp == 'I' || cTmp == 'i')  evesetvalue[profile][num] = DAMAGEICE;
    if (cTmp == 'Z' || cTmp == 'z')  evesetvalue[profile][num] = DAMAGEZAP;
    num++;
    goto_colon(fileread);  cTmp = get_first_letter(fileread);
    evesetyesno[profile][num] = (cTmp == 'T' || cTmp == 't');
    fscanf(fileread, "%d", &iTmp);  evesetvalue[profile][num] = iTmp;
    num++;
    goto_colon(fileread);  cTmp = get_first_letter(fileread);
    evesetyesno[profile][num] = (cTmp == 'T' || cTmp == 't');
    fscanf(fileread, "%d", &iTmp);  evesetvalue[profile][num] = iTmp;
    num++;
    goto_colon(fileread);  cTmp = get_first_letter(fileread);
    evesetyesno[profile][num] = (cTmp == 'T' || cTmp == 't');
    fscanf(fileread, "%d", &iTmp);  evesetvalue[profile][num] = iTmp;
    num++;
    goto_colon(fileread);  cTmp = get_first_letter(fileread);
    evesetyesno[profile][num] = (cTmp == 'T' || cTmp == 't');
    cTmp = get_first_letter(fileread);  iTmp = 0;
    if (cTmp == 'T') iTmp = DAMAGEINVERT;
    if (cTmp == 'C') iTmp = DAMAGECHARGE;
    fscanf(fileread, "%d", &tTmp);  evesetvalue[profile][num] = iTmp | tTmp;
    num++;
    goto_colon(fileread);  cTmp = get_first_letter(fileread);
    evesetyesno[profile][num] = (cTmp == 'T' || cTmp == 't');
    cTmp = get_first_letter(fileread);  iTmp = 0;
    if (cTmp == 'T') iTmp = DAMAGEINVERT;
    if (cTmp == 'C') iTmp = DAMAGECHARGE;
    fscanf(fileread, "%d", &tTmp);  evesetvalue[profile][num] = iTmp | tTmp;
    num++;
    goto_colon(fileread);  cTmp = get_first_letter(fileread);
    evesetyesno[profile][num] = (cTmp == 'T' || cTmp == 't');
    cTmp = get_first_letter(fileread);  iTmp = 0;
    if (cTmp == 'T') iTmp = DAMAGEINVERT;
    if (cTmp == 'C') iTmp = DAMAGECHARGE;
    fscanf(fileread, "%d", &tTmp);  evesetvalue[profile][num] = iTmp | tTmp;
    num++;
    goto_colon(fileread);  cTmp = get_first_letter(fileread);
    evesetyesno[profile][num] = (cTmp == 'T' || cTmp == 't');
    cTmp = get_first_letter(fileread);  iTmp = 0;
    if (cTmp == 'T') iTmp = DAMAGEINVERT;
    if (cTmp == 'C') iTmp = DAMAGECHARGE;
    fscanf(fileread, "%d", &tTmp);  evesetvalue[profile][num] = iTmp | tTmp;
    num++;
    goto_colon(fileread);  cTmp = get_first_letter(fileread);
    evesetyesno[profile][num] = (cTmp == 'T' || cTmp == 't');
    cTmp = get_first_letter(fileread);  iTmp = 0;
    if (cTmp == 'T') iTmp = DAMAGEINVERT;
    if (cTmp == 'C') iTmp = DAMAGECHARGE;
    fscanf(fileread, "%d", &tTmp);  evesetvalue[profile][num] = iTmp | tTmp;
    num++;
    goto_colon(fileread);  cTmp = get_first_letter(fileread);
    evesetyesno[profile][num] = (cTmp == 'T' || cTmp == 't');
    cTmp = get_first_letter(fileread);  iTmp = 0;
    if (cTmp == 'T') iTmp = DAMAGEINVERT;
    if (cTmp == 'C') iTmp = DAMAGECHARGE;
    fscanf(fileread, "%d", &tTmp);  evesetvalue[profile][num] = iTmp | tTmp;
    num++;
    goto_colon(fileread);  cTmp = get_first_letter(fileread);
    evesetyesno[profile][num] = (cTmp == 'T' || cTmp == 't');
    cTmp = get_first_letter(fileread);  iTmp = 0;
    if (cTmp == 'T') iTmp = DAMAGEINVERT;
    if (cTmp == 'C') iTmp = DAMAGECHARGE;
    fscanf(fileread, "%d", &tTmp);  evesetvalue[profile][num] = iTmp | tTmp;
    num++;
    goto_colon(fileread);  cTmp = get_first_letter(fileread);
    evesetyesno[profile][num] = (cTmp == 'T' || cTmp == 't');
    cTmp = get_first_letter(fileread);  iTmp = 0;
    if (cTmp == 'T') iTmp = DAMAGEINVERT;
    if (cTmp == 'C') iTmp = DAMAGECHARGE;
    fscanf(fileread, "%d", &tTmp);  evesetvalue[profile][num] = iTmp | tTmp;
    num++;
    goto_colon(fileread);  cTmp = get_first_letter(fileread);
    evesetyesno[profile][num] = (cTmp == 'T' || cTmp == 't');
    fscanf(fileread, "%d", &iTmp);  evesetvalue[profile][num] = iTmp;
    num++;
    goto_colon(fileread);  cTmp = get_first_letter(fileread);
    evesetyesno[profile][num] = (cTmp == 'T' || cTmp == 't');
    fscanf(fileread, "%d", &iTmp);  evesetvalue[profile][num] = iTmp;
    num++;
    goto_colon(fileread);  cTmp = get_first_letter(fileread);
    evesetyesno[profile][num] = (cTmp == 'T' || cTmp == 't');
    fscanf(fileread, "%d", &iTmp);  evesetvalue[profile][num] = iTmp;
    num++;
    goto_colon(fileread);  cTmp = get_first_letter(fileread);
    evesetyesno[profile][num] = (cTmp == 'T' || cTmp == 't');
    fscanf(fileread, "%d", &iTmp);  evesetvalue[profile][num] = iTmp;
    num++;
    goto_colon(fileread);  cTmp = get_first_letter(fileread);
    evesetyesno[profile][num] = (cTmp == 'T' || cTmp == 't');
    fscanf(fileread, "%d", &iTmp);  evesetvalue[profile][num] = iTmp;
    num++;
    goto_colon(fileread);  cTmp = get_first_letter(fileread);
    evesetyesno[profile][num] = (cTmp == 'T' || cTmp == 't');
    cTmp = get_first_letter(fileread);
    evesetvalue[profile][num] = (cTmp == 'T' || cTmp == 't');
    num++;
    goto_colon(fileread);  cTmp = get_first_letter(fileread);
    evesetyesno[profile][num] = (cTmp == 'T' || cTmp == 't');
    cTmp = get_first_letter(fileread);
    evesetvalue[profile][num] = (cTmp == 'T' || cTmp == 't');
    num++;
    goto_colon(fileread);  cTmp = get_first_letter(fileread);
    evesetyesno[profile][num] = (cTmp == 'T' || cTmp == 't');
    cTmp = get_first_letter(fileread);
    evesetvalue[profile][num] = MISNORMAL;
    if (cTmp == 'R' || cTmp == 'r')  evesetvalue[profile][num] = MISREFLECT;
    if (cTmp == 'D' || cTmp == 'd')  evesetvalue[profile][num] = MISDEFLECT;
    num++;
    goto_colon(fileread);  cTmp = get_first_letter(fileread);
    evesetyesno[profile][num] = (cTmp == 'T' || cTmp == 't');
    fscanf(fileread, "%f", &fTmp);  fTmp *= 16;
    evesetvalue[profile][num] = fTmp;
    num++;
    goto_colon(fileread);  cTmp = get_first_letter(fileread);
    evesetyesno[profile][num] = (cTmp == 'T' || cTmp == 't');
    evesetvalue[profile][num] = btrue;
    num++;
    goto_colon(fileread);  cTmp = get_first_letter(fileread);
    evesetyesno[profile][num] = (cTmp == 'T' || cTmp == 't');
    evesetvalue[profile][num] = btrue;
    num++;


    // Now read in the add values
    num = 0;
    goto_colon(fileread);  fscanf(fileread, "%f", &fTmp);		//Jump Power
    eveaddvalue[profile][num] = fTmp * 16;
    num++;
    goto_colon(fileread);  fscanf(fileread, "%f", &fTmp);		//Bump dampen change
    eveaddvalue[profile][num] = fTmp * 127;
    num++;
    goto_colon(fileread);  fscanf(fileread, "%f", &fTmp);		//Bounciness change
    eveaddvalue[profile][num] = fTmp * 127;
    num++;
    goto_colon(fileread);  fscanf(fileread, "%f", &fTmp);
    eveaddvalue[profile][num] = fTmp * 4;
    num++;
    goto_colon(fileread);  fscanf(fileread, "%f", &fTmp);
    eveaddvalue[profile][num] = fTmp * 127;
    num++;
    goto_colon(fileread);  fscanf(fileread, "%d", &iTmp);
    eveaddvalue[profile][num] = iTmp;
    num++;
    goto_colon(fileread);  fscanf(fileread, "%d", &iTmp);
    eveaddvalue[profile][num] = iTmp;
    num++;
    goto_colon(fileread);  fscanf(fileread, "%d", &iTmp);
    eveaddvalue[profile][num] = iTmp;
    num++;
    goto_colon(fileread);  fscanf(fileread, "%d", &iTmp);
    eveaddvalue[profile][num] = iTmp;
    num++;
    goto_colon(fileread);  fscanf(fileread, "%d", &iTmp);  // Defense is backwards
    eveaddvalue[profile][num] = -iTmp;
    num++;
    goto_colon(fileread);  fscanf(fileread, "%f", &fTmp);
    eveaddvalue[profile][num] = fTmp * 4;
    num++;
    goto_colon(fileread);  fscanf(fileread, "%f", &fTmp);
    eveaddvalue[profile][num] = fTmp * 4;
    num++;
    goto_colon(fileread);  fscanf(fileread, "%f", &fTmp);
    eveaddvalue[profile][num] = fTmp * 4;
    num++;
    goto_colon(fileread);  fscanf(fileread, "%f", &fTmp);
    eveaddvalue[profile][num] = fTmp * 4;
    num++;
    goto_colon(fileread);  fscanf(fileread, "%f", &fTmp);
    eveaddvalue[profile][num] = fTmp * 4;
    num++;
    goto_colon(fileread);  fscanf(fileread, "%f", &fTmp);
    eveaddvalue[profile][num] = fTmp * 4;
    num++;

    // Clear expansions...
    evecontspawntime[profile] = 0;
    evecontspawnamount[profile] = 0;
    evecontspawnfacingadd[profile] = 0;
    evecontspawnpip[profile] = 0;
    evewaveindex[profile] = NULL;
    evestayifnoowner[profile] = 0;
    eveoverlay[profile] = 0;
    // Read expansions
    while (goto_colon_yesno(fileread))
    {
      idsz = get_idsz(fileread);
      fscanf(fileread, "%c%d", &cTmp, &iTmp);

      if (MAKE_IDSZ("AMOU") == idsz)  evecontspawnamount[profile] = iTmp;
      if (MAKE_IDSZ("TYPE") == idsz)  evecontspawnpip[profile] = iTmp;
      if (MAKE_IDSZ("TIME") == idsz)  evecontspawntime[profile] = iTmp;
      if (MAKE_IDSZ("FACE") == idsz)  evecontspawnfacingadd[profile] = iTmp;
      if (MAKE_IDSZ("SEND") == idsz && iTmp > -1 && iTmp < MAXWAVE) evewaveindex[profile] = capwaveindex[enceve[profile]][iTmp];
      if (MAKE_IDSZ("STAY") == idsz)  evestayifnoowner[profile] = iTmp;
      if (MAKE_IDSZ("OVER") == idsz)  eveoverlay[profile] = iTmp;
    }


    // All done ( finally )
    fs_fileClose(fileread);
  }
}

//--------------------------------------------------------------------------------------------
Uint16 get_free_enchant()
{
  // ZZ> This function returns the next free enchantment or MAXENCHANT if there are none
  if (numfreeenchant > 0)
  {
    numfreeenchant--;
    return freeenchant[numfreeenchant];
  }
  return MAXENCHANT;
}

//--------------------------------------------------------------------------------------------
void unset_enchant_value(Uint16 enchantindex, Uint8 valueindex)
{
  // ZZ> This function unsets a set value
  Uint16 character;

  if (encsetyesno[enchantindex][valueindex])
  {
    character = enctarget[enchantindex];
    switch (valueindex)
    {
    case SETDAMAGETYPE:
      chrdamagetargettype[character] = encsetsave[enchantindex][valueindex];
      break;

    case SETNUMBEROFJUMPS:
      chrjumpnumberreset[character] = encsetsave[enchantindex][valueindex];
      break;

    case SETLIFEBARCOLOR:
      chrlifecolor[character] = encsetsave[enchantindex][valueindex];
      break;

    case SETMANABARCOLOR:
      chrmanacolor[character] = encsetsave[enchantindex][valueindex];
      break;

    case SETSLASHMODIFIER:
      chrdamagemodifier[character][DAMAGESLASH] = encsetsave[enchantindex][valueindex];
      break;

    case SETCRUSHMODIFIER:
      chrdamagemodifier[character][DAMAGECRUSH] = encsetsave[enchantindex][valueindex];
      break;

    case SETPOKEMODIFIER:
      chrdamagemodifier[character][DAMAGEPOKE] = encsetsave[enchantindex][valueindex];
      break;

    case SETHOLYMODIFIER:
      chrdamagemodifier[character][DAMAGEHOLY] = encsetsave[enchantindex][valueindex];
      break;

    case SETEVILMODIFIER:
      chrdamagemodifier[character][DAMAGEEVIL] = encsetsave[enchantindex][valueindex];
      break;

    case SETFIREMODIFIER:
      chrdamagemodifier[character][DAMAGEFIRE] = encsetsave[enchantindex][valueindex];
      break;

    case SETICEMODIFIER:
      chrdamagemodifier[character][DAMAGEICE] = encsetsave[enchantindex][valueindex];
      break;

    case SETZAPMODIFIER:
      chrdamagemodifier[character][DAMAGEZAP] = encsetsave[enchantindex][valueindex];
      break;

    case SETFLASHINGAND:
      chrflashand[character] = encsetsave[enchantindex][valueindex];
      break;

    case SETLIGHTBLEND:
      chrlight[character] = encsetsave[enchantindex][valueindex];
      break;

    case SETALPHABLEND:
      chralpha[character] = encsetsave[enchantindex][valueindex];
      break;

    case SETSHEEN:
      chrsheen[character] = encsetsave[enchantindex][valueindex];
      break;

    case SETFLYTOHEIGHT:
      chrflyheight[character] = encsetsave[enchantindex][valueindex];
      break;

    case SETWALKONWATER:
      chrwaterwalk[character] = encsetsave[enchantindex][valueindex];
      break;

    case SETCANSEEINVISIBLE:
      chrcanseeinvisible[character] = encsetsave[enchantindex][valueindex];
      break;

    case SETMISSILETREATMENT:
      chrmissiletreatment[character] = encsetsave[enchantindex][valueindex];
      break;

    case SETCOSTFOREACHMISSILE:
      chrmissilecost[character] = encsetsave[enchantindex][valueindex];
      chrmissilehandler[character] = character;
      break;

    case SETMORPH:
      // Need special handler for when this is removed
      change_character(character, chrbasemodel[character], encsetsave[enchantindex][valueindex], LEAVEALL);
      break;

    case SETCHANNEL:
      chrcanchannel[character] = encsetsave[enchantindex][valueindex];
      break;

    }
    encsetyesno[enchantindex][valueindex] = bfalse;
  }
}

//--------------------------------------------------------------------------------------------
void remove_enchant_value(Uint16 enchantindex, Uint8 valueindex)
{
  // ZZ> This function undoes cumulative modification to character stats
  float fvaluetoadd;
  int valuetoadd;

  Uint16 character = enctarget[enchantindex];
  switch (valueindex)
  {
  case ADDJUMPPOWER:
    fvaluetoadd = encaddsave[enchantindex][valueindex] / 16.0;
    chrjump[character] -= fvaluetoadd;
    break;

  case ADDBUMPDAMPEN:
    fvaluetoadd = encaddsave[enchantindex][valueindex] / 128.0;
    chrbumpdampen[character] -= fvaluetoadd;
    break;

  case ADDBOUNCINESS:
    fvaluetoadd = encaddsave[enchantindex][valueindex] / 128.0;
    chrdampen[character] -= fvaluetoadd;
    break;

  case ADDDAMAGE:
    valuetoadd = encaddsave[enchantindex][valueindex];
    chrdamageboost[character] -= valuetoadd;
    break;

  case ADDSIZE:
    fvaluetoadd = encaddsave[enchantindex][valueindex] / 128.0;
    chrsizegoto[character] -= fvaluetoadd;
    chrsizegototime[character] = SIZETIME;
    break;

  case ADDACCEL:
    fvaluetoadd = encaddsave[enchantindex][valueindex] / 1000.0;
    chrmaxaccel[character] -= fvaluetoadd;
    break;

  case ADDRED:
    valuetoadd = encaddsave[enchantindex][valueindex];
    chrredshift[character] -= valuetoadd;
    break;

  case ADDGRN:
    valuetoadd = encaddsave[enchantindex][valueindex];
    chrgrnshift[character] -= valuetoadd;
    break;

  case ADDBLU:
    valuetoadd = encaddsave[enchantindex][valueindex];
    chrblushift[character] -= valuetoadd;
    break;

  case ADDDEFENSE:
    valuetoadd = encaddsave[enchantindex][valueindex];
    chrdefense[character] -= valuetoadd;
    break;

  case ADDMANA:
    valuetoadd = encaddsave[enchantindex][valueindex];
    chrmanamax[character] -= valuetoadd;
    chrmana[character] -= valuetoadd;
    if (chrmana[character] < 0) chrmana[character] = 0;
    break;

  case ADDLIFE:
    valuetoadd = encaddsave[enchantindex][valueindex];
    chrlifemax[character] -= valuetoadd;
    chrlife[character] -= valuetoadd;
    if (chrlife[character] < 1) chrlife[character] = 1;
    break;

  case ADDSTRENGTH:
    valuetoadd = encaddsave[enchantindex][valueindex];
    chrstrength[character] -= valuetoadd;
    break;

  case ADDWISDOM:
    valuetoadd = encaddsave[enchantindex][valueindex];
    chrwisdom[character] -= valuetoadd;
    break;

  case ADDINTELLIGENCE:
    valuetoadd = encaddsave[enchantindex][valueindex];
    chrintelligence[character] -= valuetoadd;
    break;

  case ADDDEXTERITY:
    valuetoadd = encaddsave[enchantindex][valueindex];
    chrdexterity[character] -= valuetoadd;
    break;

  }
}
